/************************************************************************************
	RF12 driver implemenation, see rf12.h for more informations

    Copyright (C) 2007 Andreas Weber (info@tech-chat.de)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

************************************************************************************/
#include "rf12.h"

#if RF12_USE_HAMMIMG_ECC == 1
prog_uint8_t hammingC[16] =
{
    0x00,   /* 0 */
    0x71,   /* 1 */
    0x62,   /* 2 */
    0x13,   /* 3 */
    0x54,   /* 4 */
    0x25,   /* 5 */
    0x36,   /* 6 */
    0x47,   /* 7 */
    0x38,   /* 8 */
    0x49,   /* 9 */
    0x5A,   /* A */
    0x2B,   /* B */
    0x6C,   /* C */
    0x1D,   /* D */
    0x0E,   /* E */
    0x7F    /* F */
};

prog_uint8_t hammingD[128] =
{
    0x00, 0x00, 0x00, 0x03, 0x00, 0x05, 0x0E, 0x07,
    0x00, 0x09, 0x0E, 0x0B, 0x0E, 0x0D, 0x0E, 0x0E,
    0x00, 0x03, 0x03, 0x03, 0x04, 0x0D, 0x06, 0x03,
    0x08, 0x0D, 0x0A, 0x03, 0x0D, 0x0D, 0x0E, 0x0D,
    0x00, 0x05, 0x02, 0x0B, 0x05, 0x05, 0x06, 0x05,
    0x08, 0x0B, 0x0B, 0x0B, 0x0C, 0x05, 0x0E, 0x0B,
    0x08, 0x01, 0x06, 0x03, 0x06, 0x05, 0x06, 0x06,
    0x08, 0x08, 0x08, 0x0B, 0x08, 0x0D, 0x06, 0x0F,
    0x00, 0x09, 0x02, 0x07, 0x04, 0x07, 0x07, 0x07,
    0x09, 0x09, 0x0A, 0x09, 0x0C, 0x09, 0x0E, 0x07,
    0x04, 0x01, 0x0A, 0x03, 0x04, 0x04, 0x04, 0x07,
    0x0A, 0x09, 0x0A, 0x0A, 0x04, 0x0D, 0x0A, 0x0F,
    0x02, 0x01, 0x02, 0x02, 0x0C, 0x05, 0x02, 0x07,
    0x0C, 0x09, 0x02, 0x0B, 0x0C, 0x0C, 0x0C, 0x0F,
    0x01, 0x01, 0x02, 0x01, 0x04, 0x01, 0x06, 0x0F,
    0x08, 0x01, 0x0A, 0x0F, 0x0C, 0x0F, 0x0F, 0x0F
};
#endif

/*
ISR(INT1_vect)	//momentan nicht verwendet
{

}

ISR(SPI_STC_vect)	//SPI übertragung fertig, nicht verwendet
{

}
*/

static void spi_run_clock () {
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
  USICR = _BV(USIWM0) | _BV(USITC);
  USICR = _BV(USIWM0) | _BV(USITC) | _BV(USICLK);
}

void rf12_cmd(uint8_t highbyte, uint8_t lowbyte)
{
    RF12_CS();
    USIDR = highbyte;
    spi_run_clock();
    USIDR = lowbyte;
    spi_run_clock();
    RF12_NOT_CS();
}

void rf12_loop_until_FFIT_RGIT(void)
{
    do
    {
        rf12_read_status_MSB();
    }
    while (!status_H.bits.RGIT_FFIT);
}

/*	rf12_read_status_MSB
	RX Mode: FFIT = The number of data bits in the RX FIFO has reached the pre-programmed limit.
					Can be cleared by any of the FIFO read methods
	TX Mode: RGIT = TX register is ready to receive the next byte
					(Can be cleared by Transmitter Register Write Command)
*/
uint8_t rf12_read_status_MSB(void)
{
    RF12_CS();
    MOSI_LOW();
    asm volatile("nop");
    if (MISO_LEVEL())
        status_H.bits.RGIT_FFIT=1;
    else
        status_H.bits.RGIT_FFIT=0;
    RF12_NOT_CS();
    return status_H.bits.RGIT_FFIT;
}

void rf12_read_status(void)
{
    RF12_CS();
    USIDR = 0x00;	//Status Read Command
    spi_run_clock();
    status_H.byte = USIDR;
    USIDR = 0x00; 	//Status Read Command
    spi_run_clock();
    status_L.byte = USIDR;
    RF12_NOT_CS();
}

void rf12_TX(uint8_t aByte)
{
    //FFIT wird gepollt um zu erkennen ob das FIFO TX
    //Register bereit ist.
    //Alternativ ist es auch möglich(wenn verbunden)
    //den Interrupt Ausgang des RF12 zu pollen: while(INT1_LEVEL());
    while (!rf12_read_status_MSB());
    rf12_cmd(RF12_TRANSMIT,aByte);
}

#if RF12_USE_HAMMIMG_ECC == 1
void rf12_TX_hamming(uint8_t aByte)
{
    while (!rf12_read_status_MSB());
    rf12_cmd(RF12_TRANSMIT,pgm_read_byte(hammingC+(aByte & 0x0F)));
    while (!rf12_read_status_MSB());
    rf12_cmd(RF12_TRANSMIT,pgm_read_byte(hammingC+(aByte>>4)));
}
#endif

#if RF12_RECEIVE_CODE
uint8_t rf12_RX(void)
{
    rf12_loop_until_FFIT_RGIT();
    RF12_CS();
    USIDR = 0xB0;
    spi_run_clock();
    USIDR = 0x00;
    spi_run_clock();
    RF12_NOT_CS();
    return USIDR;
}

uint8_t rf12_RX_hamming(void)
{
    rf12_loop_until_FFIT_RGIT();
    RF12_CS();
    USIDR = 0xB0;
    spi_run_clock();
    USIDR = 0x00;
    spi_run_clock();
    uint8_t lower_nibble=USIDR;
    RF12_NOT_CS();
    rf12_loop_until_FFIT_RGIT();
    RF12_CS();
    USIDR = 0xB0;
    spi_run_clock();
    USIDR = 0x00;
    spi_run_clock();
    RF12_NOT_CS();
    return pgm_read_byte_near(hammingD+(lower_nibble&0x7F)) | pgm_read_byte_near(hammingD+(USIDR&0x7F))<<4;
}
#endif

void rf12_send(const uint8_t* buf, uint8_t cnt)
{
    if (!cnt) return;
    rf12_TX(0xAA);  //PREAMBLE
    rf12_TX(0x2D);  //SYNC HI BYTE
    rf12_TX(0xD4);  //SYNC LOW BYTE
    //XOR mit 0xAA ist ein quick&dirty hack da der Empfänger bei Nullen
    //am Anfang wohl außer Tritt gerät

#if RF12_USE_HAMMIMG_ECC ==1
    rf12_TX_hamming(cnt^0xAA);
#else
    rf12_TX(cnt^0xAA);
#endif
    uint8_t chksum=cnt;
    while (cnt--)
    {
#if RF12_USE_HAMMIMG_ECC ==1
        rf12_TX_hamming(*buf);
#else
        rf12_TX(*buf);
#endif
        chksum+=*buf++;
    }
#if RF12_USE_HAMMIMG_ECC ==1
    rf12_TX_hamming(chksum);
#else
    rf12_TX(chksum);
#endif
    rf12_TX(0x0);	//dummy byte
}

#if RF12_RECEIVE_CODE
//returns 0 if no data is available
//returns -1(255) if there is a CRC Error
//else, returns number of received byte
uint8_t rf12_read(uint8_t* buf, const uint8_t max)
{
    uint8_t len;
#if RF12_USE_HAMMIMG_ECC ==1
    len=rf12_RX_hamming()^0xAA;
#else
    len=rf12_RX()^0xAA;
#endif
    uint8_t checksum=len;
    uint8_t received_checksum;
    uint8_t i=len;
    while (i--)
    {
#if RF12_USE_HAMMIMG_ECC ==1
        *buf=rf12_RX_hamming();
#else
        *buf=rf12_RX();
#endif
        checksum+=*buf++;
    }
#if RF12_USE_HAMMIMG_ECC ==1
    received_checksum=rf12_RX_hamming();
#else
    received_checksum=rf12_RX();
#endif
    if (received_checksum==checksum)
        return len;
    else
        return -1;
}
#endif

void rf12_init(void)
{
#if 0
    //configure SPI (Atmega32 s.136)
    //SPI CLK auf F_CPU/2
  SPCR = /*_BV(SPIE) |*/ _BV(SPE) | _BV(MSTR) /* | _BV(SPR1) | _BV(SPR0) */;
  SPSR |= _BV(SPI2X);
#else
  USICR = _BV(USIWM0); // 3-wire, software clock strobe
#endif

#if 0
    uint8_t i;
    for (i=0;i<100;++i)	//300ms für PowerUp warten
        _delay_ms(3);
#endif

    //14. Low Battery Detector and Microcontroller Clock Divider Command
    //Clock Output selected in rf12_config.h
    //2.2V Threshold
    rf12_cmd(RF12_LOW_BAT_AND_CLK_DIV, RF12_LOW(0) | RF12_DIV(RF12_CLOCK_OUTPUT));

    //1.Configuration Setting Command
    //EL(internal data register), EF(FIFO mode), 433MHz, 12pF crystal Load Capacitance
    rf12_cmd(RF12_CFG_SETTING,RF12_EL | RF12_EF | RF12_868MHz | RF12_LOAD_C(12.0));

    //6. Data Filter Command
    rf12_cmd(RF12_DATA_FILTER, RF12_AUTOLOCK | RF12_FASTMODE | RF12_DQD_THRESHOLD(4));

    //7. FIFO and Reset ModeCommand
    rf12_cmd(RF12_FIFO_RESET, RF12_FIFO_IT_LEVEL(8) | RF12_FIFO_FILL(0) | RF12_DR);

    //12. Wake-Up Timer Command
    //do not use wake-up timer
    rf12_cmd(RF12_WAKE_UP(0),0x00);

    //13. Low Duty.Cycle Command
    //do not use duty cycle
    rf12_cmd(RF12_DUTY_CYCLE(0),0x00);

    //9. AFC Command
    rf12_cmd(RF12_AFC, RF12_AFC_AUTO(2) | RF12_AFC_RANGE_LIMIT(2) | RF12_AFC_en | RF12_AFC_oe | RF12_AFC_fi);

    //3. Freqency Settings 433.92
    rf12_cmd(RF12_FREQ_SETTING_HIGH(RF12_FREQ_868Band(868.35)),RF12_FREQ_SETTING_LOW(RF12_FREQ_868Band(868.35)));

    //5. Receiver Control Command
    rf12_cmd(RF12_RECEIVER_CONTROL(RF12_VDI_MEDIUM | RF12_P20_VDI),RF12_BW_200 | RF12_LNA_0dB | RF12_RSSI_85dBm);

    //4. Data Rate Command, default 9600kbps
    //    rf12_cmd(RF12_DATA_RATE,RF12_9600);
    rf12_cmd(RF12_DATA_RATE,RF12_19200);

    //10. TX Configuration Control Command
    //!mp, freq-shift 5*15kHz, MAX OUT
    rf12_cmd(RF12_TX_CFG, 0x40);
}

